from __future__ import annotations

import datetime as _dt


class SPDXLicenseList:
    """SPDX license list."""

    def __init__(self, data: dict):
        self._data = data
        self._map = {licence["licenseId"]: licence for licence in data["licenses"]}
        return

    @property
    def licenses(self) -> list[dict]:
        """List of SPDX licenses."""
        return self._data["licenses"]

    @property
    def ids(self) -> list[str]:
        """List of SPDX license IDs."""
        return list(self._map.keys())

    @property
    def release_date(self) -> _dt.date:
        """Release date of the SPDX license list."""
        return _dt.datetime.fromisoformat(self._data["releaseDate"]).date()

    @property
    def version(self) -> str:
        """Version of the SPDX license list."""
        return self._data["licenseListVersion"]

    def get(self, key: str) -> dict | None:
        """Get a license by its key."""

        return self._map.get(key)

    def __getitem__(self, key: str) -> dict:
        return self._map[key]

    def __contains__(self, key: str) -> bool:
        return key in self._map

    def __repr__(self):
        return f"<SPDXLicenseList {self.version}>"



from __future__ import annotations as annotations

from typing import TYPE_CHECKING as TYPE_CHECKING
import json as _json

from licenseman.spdx.license import SPDXLicense as _SPDXLicense

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Sequence, Generator
    from licenseman.spdx.license_list import SPDXLicenseList


class SPDXLicenseDB:

    def __init__(
        self,
        license_list: SPDXLicenseList,
        db_path: Path,
        in_memory: bool = False,
        verify: bool = True,
        licenses: dict[str, _SPDXLicense] | None = None,
    ):
        self._license_list = license_list
        self._db_path = db_path
        self._in_memory = in_memory
        self._verify = verify
        self._licenses: dict[str, _SPDXLicense] = licenses or {}
        if in_memory:
            self.load(verify=verify)
        elif verify:
            for _ in self.get(verify=True):
                pass
        return

    def load(self, license_ids: Sequence[str] | None = None, verify: bool | None = None):
        license_ids = license_ids or self._license_list.ids
        for license_id, license_data in zip(license_ids, self.get(license_ids, verify=verify)):
            self._licenses[license_id] = license_data
        return

    def get(self, license_ids: Sequence[str] | None = None, verify: bool | None = None) -> Generator[_SPDXLicense]:
        for license_id in license_ids or self._license_list.ids:
            if license_id in self._licenses:
                yield self._licenses[license_id]
            else:
                with open(self._db_path / f"{license_id}.json") as f:
                    data = _json.load(f)
                yield _SPDXLicense(
                    data,
                    verify=verify if verify is not None else self._verify,
                )

    def alts(self, license_ids: Sequence[str] | None = None):
        license_ids = license_ids or self._license_list.ids
        alts: dict[str, list[dict[str, str]]] = {}
        for license in self.get(license_ids):
            for alt_name, alt_data in license.alts.items():
                alts.setdefault(alt_name, []).append({"id": license.id, **alt_data})
        return alts

    def __getitem__(self, license_id: str) -> _SPDXLicense:
        return self.get([license_id]).__next__()

    def __contains__(self, license_id: str) -> bool:
        return license_id in self._license_list




from licenseman.spdx.entry import SPDXEntry


class SPDXLicenseException(SPDXEntry):
    """SPDX License definition.

    Parameters
    ----------
    xml
        SPDX license XML content as a string.

    References
    ----------
    - [SPDX Docs](https://github.com/spdx/license-list-XML/blob/main/DOCS/README.md)
    - [SPDX Docs - XML Fields](https://github.com/spdx/license-list-XML/blob/main/DOCS/xml-fields.md)
    - [XML Schema](https://github.com/spdx/license-list-XML/blob/main/schema/ListedLicense.xsd)
    - [GitHub Repository](https://github.com/spdx/license-list-XML)
    """

    def __init__(self, data: dict, verify: bool = True):
        super().__init__(data=data, entry_type="exception", verify=verify)
        return

    def __repr__(self):
        return f"<SPDXLicenseException {self.id}>"



from __future__ import annotations as _annotations

import copy as _copy
from typing import TYPE_CHECKING as _TYPE_CHECKING
import datetime as _dt
from xml.etree import ElementTree as _ElementTree

from licenseman import logger as _logger
from licenseman.spdx.license_text import SPDXLicenseTextPlain

if _TYPE_CHECKING:
    from typing import Literal, Any



class SPDXEntry:
    """SPDX License or exception definition.

    References
    ----------
    - [SPDX Docs](https://github.com/spdx/license-list-XML/blob/main/DOCS/README.md)
    - [SPDX Docs - XML Fields](https://github.com/spdx/license-list-XML/blob/main/DOCS/xml-fields.md)
    - [XML Schema](https://github.com/spdx/license-list-XML/blob/main/schema/ListedLicense.xsd)
    - [GitHub Repository](https://github.com/spdx/license-list-XML)
    """

    def __init__(self, data: dict, entry_type: Literal["license", "exception"], verify: bool = True):
        try:
            root = _ElementTree.fromstring(data["xml"])
        except _ElementTree.ParseError as e:
            raise Exception(f"Error parsing license XML content.") from e
        self._ns_url = 'http://www.spdx.org/license'
        self._ns: dict = {'': self._ns_url}
        self._xml: _ElementTree.Element = root.find(entry_type, self._ns)
        self._entry_type: Literal["license", "exception"] = entry_type
        self._data: dict = data
        if verify:
            self.verify()
        return

    def verify(self):

        def log(key_json: str, missing_in: Literal["xml", "json"], data: Any, key_xml: str | None = None):
            if key_xml is None:
                key_xml = key_json
            if missing_in == "xml":
                missing_source = "XML"
                existing_source = "JSON"
                missing_key = key_xml
                existing_key = key_json
            else:
                missing_source = "JSON"
                existing_source = "XML"
                missing_key = key_json
                existing_key = key_xml
            _logger.notice(
                f"{self.id} License{" Exception" if self._entry_type == "exception" else ""} Verification",
                f"The value of '{missing_key}' is not defined in the {missing_source} data. "
                f"Using the {existing_source} data value of '{existing_key}':",
                _logger.pretty(data)
            )
            return

        def osi_approved():
            key = "isOsiApproved"
            xml_raw = self._xml.attrib.get(key)
            if xml_raw == "true":
                xml = True
            elif xml_raw == "false":
                xml = False
            else:
                if xml_raw is not None:
                    raise Exception(f"Invalid value for '{key}' in XML data: {xml_raw}")
                xml = None
            json = self._data[key]
            if json != xml:
                if xml is None:
                    log(key, "xml", json)
                    return
                if json is None:
                    log(key, "json", xml)
                    self._data[key] = xml
                    return
                raise Exception(
                    "OSI approved mismatch between XML and JSON data. "
                    f"XML: {xml}, JSON: {json}"
                )
            return

        def deprecated_version():
            key = "deprecatedVersion"
            xml = self._xml.attrib.get(key)
            json = self._data.get(key)
            if json != xml:
                if xml is None:
                    log(key_json=key, missing_in="xml", data=json)
                elif json is None:
                    log(key_json=key, missing_in="json", data=xml)
                    self._data[key] = xml
                else:
                    raise Exception(
                        "Deprecated version mismatch between XML and JSON data. "
                        f"XML: {xml}, JSON: {json}"
                    )
            return

        def cross_refs():
            xml_elem = self._xml.find('crossRefs', self._ns)
            xml = sorted(
                [ref.text.strip() for ref in xml_elem.findall('crossRef', self._ns)]
            ) if xml_elem else []
            json_seealso = sorted(self._data.get("seeAlso", []))

            if json_seealso != xml:
                if not xml:
                    log("seeAlso", "xml", data=json_seealso)
                    return
                if not json_seealso:
                    log("seeAlso", "json", data=xml, key_xml="crossRefs")
                    self._data["seeAlso"] = xml
                    return
                raise Exception(
                    "Cross references mismatch between XML and JSON data. "
                    f"XML: {xml}, JSON: {json_seealso}"
                )
            if self._entry_type == "license":
                json = sorted([ref["url"] for ref in self._data.get("crossRef", [])])
                if json != json_seealso:
                    raise Exception(
                        "Cross references mismatch between 'crossRefs' and 'seeAlso' JSON data. ",
                        f"CrossRefs: {json}, SeeAlso: {json_seealso}"
                    )
            return

        if self.id != self._xml.attrib.get('licenseId'):
            raise Exception("License ID mismatch between XML and JSON data.")
        if self._data["name"] != self._xml.attrib.get('name'):
            raise Exception("License name mismatch between XML and JSON data.")

        deprecated_version()
        cross_refs()
        if self._entry_type == "license":
            osi_approved()
        return

    def generate_text_plain(
        self,
        title: str | bool = True,
        copyright_notice: str | bool = False,
        optionals: bool | list[bool] = True,
        alts: dict[str, str] | None = None,
        line_length: int = 88,
        list_indent: int = 0,
        item_indent: int = 1,
        item_spacing: int = 1,
        bullet: str | int | None = 1,
        title_centered: bool = False,
        title_underline: Literal["-", "=", "_", "*"] = "=",
        title_underline_full: bool = False,
        subtitle_underline: Literal["-", "=", "_", "*"] = "-",
        line_breaks: int = 2,
    ) -> str:
        return SPDXLicenseTextPlain(text=self.text_xml).generate(
            title=title,
            copyright_notice=copyright_notice,
            optionals=optionals,
            alts=alts,
            line_length=line_length,
            list_indent=list_indent,
            item_indent=item_indent,
            item_spacing=item_spacing,
            bullet=bullet,
            title_centered=title_centered,
            title_underline=title_underline,
            title_underline_full=title_underline_full,
            subtitle_underline=subtitle_underline,
            line_breaks=line_breaks,
        )

    @property
    def raw_data(self) -> dict:
        """Raw license data."""
        return self._data

    @property
    def id(self) -> str:
        """SPDX license ID."""
        return self._data["licenseId" if self._entry_type == "license" else "licenseExceptionId"]

    @property
    def name(self) -> str:
        """Full name of the license"""
        return self._data["name"]

    @property
    def text_plain(self) -> str:
        """Original license text in plain text format."""
        return self._data["licenseText" if self._entry_type == "license" else "licenseExceptionText"]

    @property
    def text_template(self) -> str | None:
        """License text template."""
        return self._data.get("standardLicenseTemplate" if self._entry_type == "license" else "licenseExceptionTemplate")

    @property
    def text_html(self) -> str | None:
        """Original license text in HTML format."""
        return self._data.get("licenseTextHtml" if self._entry_type == "license" else "exceptionTextHtml")

    @property
    def text_xml(self) -> _ElementTree.Element:
        return self._xml.find('text', self._ns)

    @property
    def text_xml_str(self) -> str:
        return self._xml_str(self.text_xml)

    @property
    def title_text_xml(self) -> _ElementTree.Element | None:
        """Title of the license as defined in the text, if any."""
        return self._xml.find('.//titleText', self._ns)

    @property
    def copyright_text_xml(self) -> _ElementTree.Element | None:
        """Copyright notice of the license is defined in the text, if any."""
        return self._xml.find('.//copyrightText', self._ns)

    @property
    def optionals_xml(self) -> list[_ElementTree.Element]:
        """Optional fields in the license text, if any."""
        return self._xml.findall('.//optional', self._ns)

    @property
    def optionals_xml_str(self) -> list[str]:
        """Optional fields in the license text, if any."""
        out = []
        for optional in self.optionals_xml:
            out.append(self._xml_str(optional))
        return out

    @property
    def alts(self) -> dict[str, dict[str, str]]:
        """

        Returns
        -------
        A dictionary where keys are the alternative field names, and values are dictionaries with keys:
        `text` : str

            Default value.
        `match` : str

            Regular expression (RegEx) pattern to validate user input for `text`.
        """
        alts = {}
        for alt in self._xml.findall('.//alt', self._ns):
            alts[alt.attrib['name']] = {'text': alt.text, 'match': alt.attrib['match']}
        return alts

    @property
    def reference_number(self) -> int:
        """Reference number of the license."""
        return self._data["referenceNumber"]

    @property
    def url_reference(self) -> str:
        """URL to the license reference page at SPDX.org."""
        return self._data["reference"]

    @property
    def url_json(self) -> str:
        """URL to the license JSON data."""
        return self._data["detailsUrl"]

    @property
    def url_cross_refs(self) -> list[str]:
        """URLs to license resources, if any."""
        return self._data.get("seeAlso", [])

    @property
    def deprecated(self) -> bool:
        """Whether the license is deprecated.

        Returns
        -------
        A boolean, or `None` if the value is not defined in the data.
        """
        return self._data["isDeprecatedLicenseId"]

    @property
    def version_deprecated(self) -> str | None:
        """Version of the SPDX License List in which the license was deprecated, if applicable.

        Returns
        -------
        Version number string, or `None` if the value is not defined in the data.
        """
        return self._data.get("deprecatedVersion")

    @property
    def obsoleted_by(self) -> list[dict[str, str]] | None:
        """New licenses that obsolete this license, if any.

        Returns
        -------
        A list of dictionaries with keys:
        `id` : str

             SPDX license ID of the successor license.
        `expression` : str

             [SPDX license expression](https://spdx.github.io/spdx-spec/v3.0.1/annexes/spdx-license-expressions/)
             which is obsoleted by the successor license;
             in most cases, this is the same as the current license's ID, unless the current license
             is a complex expression, and only a part of it is obsoleted by the successor.
        """
        return [
            {"id": elem.text, "expression": elem.attrib.get("expression")}
            for elem in self._xml.findall('.//obsoletedBy', self._ns)
        ]

    @property
    def version_added(self) -> str | None:
        """Version of the SPDX License List in which the license was first added.

        Returns
        -------
        Version number string, or `None` if the value is not defined in the data.
        """
        return self._xml.attrib.get('listVersionAdded')

    @property
    def comments(self) -> str | None:
        """Comments about the license, if any."""
        return self._data.get("licenseComments")

    @property
    def notes(self) -> str | None:
        """General comments about the entry, if any."""
        elem = self._xml.find('notes', self._ns)
        return elem.text if elem is not None else None

    @property
    def xml(self) -> _ElementTree.Element:
        return self._xml

    @property
    def xml_attributes(self) -> dict[str, str]:
        return self._xml.attrib

    @property
    def xml_tags(self) -> list[str]:
        """Set of all XML tags used in the license."""
        def traverse(elem):
            tags.add(elem.tag.removeprefix(f"{{{self._ns_url}}}"))
            for child in elem:
                traverse(child)
            return

        tags = set()
        traverse(self._xml)
        return list(tags)

    def xml_tag_paths(self, tag: str) -> list[str]:
        """Get all paths to XML elements with a specific tag."""

        def find_paths(current_element, current_path):
            # Construct the current element's path
            current_tag = current_element.tag.removeprefix(f"{{{self._ns_url}}}")
            new_path = f"{current_path}/{current_tag}" if current_path else current_tag
            if current_tag == tag:
                paths.append(new_path)
            for child in current_element:
                find_paths(child, new_path)
            return

        paths = []
        find_paths(self._xml, "")
        return paths

    def __repr__(self):
        return f"<SPDXEntry {self.id}>"

    def __str__(self):
        return self.text_plain

    @staticmethod
    def _xml_str(element: _ElementTree.Element):
        optional_copy = _copy.deepcopy(element)
        optional_copy.tail = None
        return _ElementTree.tostring(
            optional_copy,
            encoding='unicode',
            xml_declaration=True,
        )


"""Functions for retrieving package data files."""

from __future__ import annotations as _annotations

from pathlib import Path as _Path

import pkgdata as _pkgdata
import pyserials as _ps

__all__ = ["get_filepath"]


def get_filepath(relative_path: str) -> _Path:
    """Get the absolute path to a package data file.

    Parameters
    ----------
    relative_path
        Path to the file relative to the package's data directory.
    """
    path_data_dir = _Path(_pkgdata.get_package_path_from_caller(top_level=False))
    filepath = path_data_dir / relative_path
    if not filepath.is_file():
        from licenseman.exception.data import DataFileNotFoundError

        raise DataFileNotFoundError(
            path_relative=relative_path,
            path_absolute=filepath,
        )
    return filepath


def spdx_to_trove_mapping() -> dict[str, str]:
    """Get the SPDX to Trove classifier mapping."""
    rel_path = "spdx/trove_classifiers.yaml"
    abs_path = get_filepath(rel_path)
    return _ps.read.yaml_from_file(abs_path)



"""Process SPDX [license expressions](https://spdx.github.io/spdx-spec/v3.0.1/annexes/spdx-license-expressions/).

See Also
--------
- https://pypi.org/project/license-expression/
"""

import re as _re

from licenseman import spdx as _spdx


def ids(expression: str) -> tuple[list[str], list[str]]:
    """Get all SPDX license and exception IDs from an expression.

    Parameters
    ----------
    expression
        SPDX license expression.

    Returns
    -------
    List of registered and custom SPDX license and exception IDs in the expression.
    """
    license_ids, license_ids_custom = license_ids(expression)
    exception_ids, exception_ids_custom = exception_ids(expression)
    return license_ids + exception_ids, license_ids_custom + exception_ids_custom


def license_ids(expression: str) -> tuple[list[str], list[str]]:
    """Get all SPDX license IDs from an expression.

    Parameters
    ----------
    expression
        SPDX license expression.

    Returns
    -------
    List of registered and custom SPDX license IDs in the expression.
    """
    return _get_ids(expression, exception=False)


def exception_ids(expression: str) -> tuple[list[str], list[str]]:
    """Get all SPDX license exception IDs from an expression.

    Parameters
    ----------
    expression
        SPDX license expression.

    Returns
    -------
    List of registered and custom SPDX license exception IDs in the expression.
    """
    return _get_ids(expression, exception=True)


def _get_ids(expression: str, exception: bool):
    list_ = _spdx._get_global_exception_list() if exception else _spdx._get_global_license_list()
    ids = sorted(list_.ids, key=len, reverse=True)  # Sort by length to match longest IDs first, e.g. 'GPL-3.0-only' before 'GPL-3.0'
    registered = _re.findall(
        rf"({'|'.join(_re.escape(exception_id) for exception_id in ids)})",
        expression
    )
    customs = _re.findall(
        rf"(?:DocumentRef-[a-zA-Z0-9-.]+:)?{'AdditionRef' if exception else 'LicenseRef'}-[a-zA-Z0-9.-]+",
        expression
    )
    return registered, customs



from __future__ import annotations as _annotations
from typing import TYPE_CHECKING as _TYPE_CHECKING
from pathlib import Path as _Path
import json as _json

import platformdirs as _platdir
import pylinks as _pl
from pylinks.exception.api import WebAPIError as _WebAPIError

from licenseman import data as _data
from licenseman.spdx.exception import SPDXLicenseException
from licenseman.spdx.license_db import SPDXLicenseDB
from licenseman.spdx.exception_db import SPDXLicenseExceptionDB
from licenseman.spdx.license_list import SPDXLicenseList
from licenseman.spdx.exception_list import SPDXExceptionList
from licenseman.spdx.license import SPDXLicense
from licenseman.spdx import expression
from licenseman import logger

if _TYPE_CHECKING:
    from typing import Literal


URL_TEMPLATE_LICENSE_XML = "https://raw.githubusercontent.com/spdx/license-list-data/refs/heads/main/license-list-XML/{}.xml"
URL_TEMPLATE_LICENSE_JSON = "https://raw.githubusercontent.com/spdx/license-list-data/refs/heads/main/json/details/{}.json"

URL_TEMPLATE_EXCEPTION_XML = "https://raw.githubusercontent.com/spdx/license-list-data/refs/heads/main/license-list-XML/exceptions/{}.xml"
URL_TEMPLATE_EXCEPTION_JSON = "https://raw.githubusercontent.com/spdx/license-list-data/refs/heads/main/json/exceptions/{}.json"

URL_LICENSE_LIST = "https://spdx.org/licenses/licenses.json"
URL_EXCEPTION_LIST = "https://spdx.org/licenses/exceptions.json"


def license_db(
    path: str | _Path = _platdir.site_cache_path(
        appauthor="RepoDynamics",
        appname="LicenseMan",
    ) / "SPDX_DB" / "licenses",
    force_update: bool = False,
    verify: bool = True,
    in_memory: bool = False,
) -> SPDXLicenseDB:
    return _db(
        typ="license",
        path=_Path(path),
        force_update=force_update,
        verify=verify,
        in_memory=in_memory,
    )


def exception_db(
    path: str | _Path = _platdir.site_cache_path(
        appauthor="RepoDynamics",
        appname="LicenseMan",
    ) / "SPDX_DB" / "exceptions",
    force_update: bool = False,
    verify: bool = True,
    in_memory: bool = False,
) -> SPDXLicenseExceptionDB:
    return _db(
        typ="exception",
        path=_Path(path),
        force_update=force_update,
        verify=verify,
        in_memory=in_memory,
    )


def license_list() -> SPDXLicenseList:
    """Get the latest version of the [SPDX license list](https://spdx.org/licenses/) from SPDX website."""
    data = _pl.http.request(URL_LICENSE_LIST, response_type="json")
    return SPDXLicenseList(data)


def exception_list() -> SPDXExceptionList:
    """Get the latest version of the [SPDX exception list](https://spdx.org/licenses/exceptions-index.html) from SPDX website."""
    data = _pl.http.request(URL_EXCEPTION_LIST, response_type="json")
    return SPDXExceptionList(data)


def license(license_id: str, verify: bool = True) -> SPDXLicense:
    """Get an SPDX license.

    Parameters
    ----------
    license_id
        SPDX license ID, e.g., 'MIT', 'GPL-2.0-or-later'.
    """
    return _license(license_id, "license", verify)


def exception(exception_id: str, verify: bool = True) -> SPDXLicenseException:
    """Get an SPDX license exception.

    Parameters
    ----------
    exception_id
        SPDX license exception ID, e.g., 'Autoconf-exception-2.0'.
    """
    return _license(exception_id, "exception", verify)


def license_xml(license_id: str) -> str:
    """Get an SPDX license definition in XML format from SPDX
    [license-list-data](https://github.com/spdx/license-list-data) repository.

    Parameters
    ----------
    license_id
        SPDX license ID, e.g., 'MIT', 'GPL-2.0-or-later'.
    """
    return _download(license_id, format="xml")


def license_json(license_id: str) -> dict:
    """Get an SPDX license definition in XML format from SPDX
    [license-list-data](https://github.com/spdx/license-list-data) repository.

    Parameters
    ----------
    license_id
        SPDX license ID, e.g., 'MIT', 'GPL-2.0-or-later'.
    """
    return _download(license_id, format="json")


def exception_xml(exception_id: str) -> str:
    """Get an SPDX license exception definition in XML format from SPDX
    [license-list-data](https://github.com/spdx/license-list-data) repository.

    Parameters
    ----------
    exception_id
        SPDX license exception ID, e.g., 'Autoconf-exception-2.0'.
    """
    return _download(exception_id, format="xml", exception=True)


def exception_json(exception_id: str) -> dict:
    """Get an SPDX license exception definition in XML format from SPDX
    [license-list-data](https://github.com/spdx/license-list-data) repository.

    Parameters
    ----------
    exception_id
        SPDX license exception ID, e.g., 'Autoconf-exception-2.0'.
    """
    return _download(exception_id, format="json", exception=True)


def trove_classifier(license_id: str) -> str | None:
    """Get the Trove classifier for an SPDX license.

    Parameters
    ----------
    license_id
        SPDX license ID, e.g., 'MIT', 'GPL-2.0-or-later'.
    """
    return _get_global_trove_mapping().get(license_id)


def _db(
    typ: Literal["license", "exception"],
    path: _Path,
    force_update: bool = False,
    verify: bool = True,
    in_memory: bool = False,
) -> SPDXLicenseDB | SPDXLicenseExceptionDB:
    if typ == "license":
        name = "license"
        class_ = SPDXLicenseDB
        func = license
        list_ = _get_global_license_list()
    else:
        name = "license exception"
        class_ = SPDXLicenseExceptionDB
        func = exception
        list_ = _get_global_exception_list()

    name_title = name.title()
    ids = list_.ids
    if force_update or not path.is_dir():
        missing_ids = ids
        intro = "Force update is enabled" if force_update else f"SPDX {name} database not found at {path}"
        logger.log(
            "info" if force_update else "notice",
            f"SPDX {name_title} Database Load",
            f"{intro}; downloading all latest SPDX {name} data."
        )
    else:
        missing_ids = []
        for license_id in ids:
            if not (path / f"{license_id}.json").is_file():
                missing_ids.append(license_id)
        if not missing_ids:
            logger.success(
                f"SPDX {name_title} Database Load",
                f"Loaded database from {path}; all {len(ids)} {name}s files found."
            )
            return class_(
                list_,
                path,
                in_memory,
                verify,
            )
        num_missing = len(missing_ids)
        num_available = len(ids) - num_missing
        logger.log(
            "notice",
            f"SPDX {name_title} Database Load",
            f"Loaded database from {path}; "
            f"found {num_missing} missing license files (available: {num_available})."
        )
    path.mkdir(parents=True, exist_ok=True)
    licenses = {}
    for missing_id in missing_ids:
        output_path = path / f"{missing_id}.json"
        data = func(missing_id, verify=False if in_memory else verify)
        with open(output_path, "w") as f:
            _json.dump(data.raw_data, f)
        logger.success(
            f"SPDX {name_title} Database Update",
            f"Downloaded '{missing_id}' to 'file://{output_path}'.",
        )
        if in_memory:
            licenses[missing_id] = data
    return class_(
        list_,
        path,
        in_memory,
        verify,
        licenses,
    )


def _license(spdx_id: str, typ: Literal["license", "exception"], verify: bool = True) -> SPDXLicense | SPDXLicenseException:
    """Get an SPDX license or exception.

    Parameters
    ----------
    SPDX_id
        SPDX ID, e.g., 'MIT', 'GPL-2.0-or-later'.
    """
    if typ == "license":
        name = "license"
        func_json = license_json
        func_xml = license_xml
        class_ = SPDXLicense
        list_ = _get_global_license_list()
    else:
        name = "license exception"
        func_json = exception_json
        func_xml = exception_xml
        class_ = SPDXLicenseException
        list_ = _get_global_exception_list()

    data = func_json(spdx_id)
    data["xml"] = func_xml(spdx_id)
    name_title = name.title()

    for list_entry_key, list_entry_val in list_[spdx_id].items():
        # 'detailsUrl', 'reference', 'referenceNumber' are not present in JSON data
        if list_entry_key not in data:
            data[list_entry_key] = list_entry_val
            logger.info(
                f"SPDX JSON {name_title} Load",
                f"Added missing '{list_entry_key}' entry to '{spdx_id}' JSON data from {name} list."
            )
        elif data[list_entry_key] != list_entry_val:
            logger.warning(
                f"SPDX JSON {name_title} Load",
                f"Mismatched '{list_entry_key}' entry in '{spdx_id}' JSON data.",
                "JSON content:",
                logger.pretty(data[list_entry_key]),
                f"{name.capitalize()} list content:",
                logger.pretty(list_entry_val),
            )
    return class_(data, verify=verify)


def _download(
    spdx_id: str,
    format: Literal["xml", "json"],
    exception: bool = False,
) -> str | dict:
    if exception:
        url = URL_TEMPLATE_EXCEPTION_XML if format == "xml" else URL_TEMPLATE_EXCEPTION_JSON
    else:
        url = URL_TEMPLATE_LICENSE_XML if format == "xml" else URL_TEMPLATE_LICENSE_JSON
    try:
        data = _pl.http.request(
            url.format(spdx_id),
            response_type="str" if format == "xml" else "json"
        )
    except _WebAPIError as e:
        msg_typ = "license" if not exception else "license exception"
        msg_format = "XML" if format == "xml" else "JSON"
        msg = f"Error downloading {msg_typ} {msg_format} for ID '{spdx_id}"
        raise Exception(msg) from e
    return data


def _get_global_license_list() -> SPDXLicenseList:
    global _LICENSE_LIST
    if _LICENSE_LIST is None:
        _LICENSE_LIST = license_list()
    return _LICENSE_LIST


def _get_global_exception_list() -> SPDXExceptionList:
    global _EXCEPTION_LIST
    if _EXCEPTION_LIST is None:
        _EXCEPTION_LIST = exception_list()
    return _EXCEPTION_LIST


def _get_global_trove_mapping() -> dict[str, str]:
    global _TROVE_MAPPING
    if _TROVE_MAPPING is None:
        _TROVE_MAPPING = _data.spdx_to_trove_mapping()["map"]
    return _TROVE_MAPPING


_LICENSE_LIST: SPDXLicenseList | None = None
_EXCEPTION_LIST: SPDXExceptionList | None = None
_TROVE_MAPPING: dict[str, str] | None = None


from __future__ import annotations as _annotations

from typing import TYPE_CHECKING as _TYPE_CHECKING
import datetime as _dt
from xml.etree import ElementTree as _ElementTree
from dataclasses import dataclass as _dataclass

from licenseman.spdx.license_text import SPDXLicenseTextPlain
from licenseman.spdx.entry import SPDXEntry

if _TYPE_CHECKING:
    from typing import Literal


@_dataclass
class SPDXLicenseCrossRef:
    """SPDX License cross reference."""
    url: str
    order: int
    timestamp: _dt.datetime
    match: str
    valid: bool
    live: bool
    wayback: bool


class SPDXLicense(SPDXEntry):
    """SPDX License definition.

    Parameters
    ----------
    xml
        SPDX license XML content as a string.

    References
    ----------
    - [SPDX Docs](https://github.com/spdx/license-list-XML/blob/main/DOCS/README.md)
    - [SPDX Docs - XML Fields](https://github.com/spdx/license-list-XML/blob/main/DOCS/xml-fields.md)
    - [XML Schema](https://github.com/spdx/license-list-XML/blob/main/schema/ListedLicense.xsd)
    - [GitHub Repository](https://github.com/spdx/license-list-XML)
    """

    def __init__(self, data: dict, verify: bool = True):
        super().__init__(data=data, entry_type="license", verify=verify)
        return

    def generate_header_plain(
        self,
        title: str | bool = True,
        copyright_notice: str | bool = False,
        optionals: bool | list[bool] = True,
        alts: dict[str, str] | None = None,
        line_length: int = 88,
        list_indent: int = 0,
        item_indent: int = 1,
        item_spacing: int = 1,
        bullet: str | int | None = 1,
        title_centered: bool = False,
        title_underline: Literal["-", "=", "_", "*"] = "=",
        title_underline_full: bool = False,
        subtitle_underline: Literal["-", "=", "_", "*"] = "-",
        line_breaks: int = 2,
    ) -> str:
        if not self.header_xml:
            return ""
        return SPDXLicenseTextPlain(text=self.header_xml).generate(
            title=title,
            copyright_notice=copyright_notice,
            optionals=optionals,
            alts=alts,
            line_length=line_length,
            list_indent=list_indent,
            item_indent=item_indent,
            item_spacing=item_spacing,
            bullet=bullet,
            title_centered=title_centered,
            title_underline=title_underline,
            title_underline_full=title_underline_full,
            subtitle_underline=subtitle_underline,
            line_breaks=line_breaks,
        )

    @property
    def header_plain(self) -> str | None:
        """Original license header in plain text format."""
        return self._data.get("standardLicenseHeader")

    @property
    def header_template(self) -> str | None:
        """License header template."""
        return self._data.get("standardLicenseHeaderTemplate")

    @property
    def header_html(self) -> str | None:
        """Original license header in HTML format."""
        return self._data.get("standardLicenseHeaderHtml")

    @property
    def header_xml(self) -> _ElementTree.Element | None:
        return self._xml.find('.//standardLicenseHeader', self._ns)

    @property
    def header_xml_str(self) -> str | None:
        return self._xml_str(self.header_xml) if self.header_xml else None

    @property
    def cross_refs(self) -> list[SPDXLicenseCrossRef]:
        """URLs to license resources, if any."""
        return [
            SPDXLicenseCrossRef(
                url=ref["url"],
                order=ref["order"],
                timestamp=_dt.datetime.strptime(ref["timestamp"], "%Y-%m-%dT%H:%M:%SZ"),
                match=ref["match"],
                valid=ref["isValid"],
                live=ref["isLive"],
                wayback=ref["isWayBackLink"]
            ) for ref in self._data.get("crossRef", [])
        ]

    @property
    def osi_approved(self) -> bool:
        """Whether the license is OSI approved.

        Returns
        -------
        A boolean, or `None` if the value is not defined in the data.
        """
        return self._data["isOsiApproved"]

    @property
    def fsf_libre(self) -> bool | None:
        """Whether the license is FSF approved.

        Returns
        -------
        A boolean, or `None` if the value is not defined in the data.
        """
        return self._data.get("isFsfLibre")

    def __repr__(self):
        return f"<SPDXLicense {self.id}>"



from __future__ import annotations as _annotations

from typing import TYPE_CHECKING as _TYPE_CHECKING
import re as _re
from xml.etree import ElementTree as ET
from textwrap import TextWrapper as _TextWrapper

import mdit as _mdit

if _TYPE_CHECKING:
    from typing import Any, Literal


class SPDXLicenseText:
    """Base text generator for SPDX licenses.

    This parses the <text> element from an SPDX license XML.
    Subclasses should implement the missing methods to generate the full text and header.

    Parameters
    ----------
    text : xml.etree.ElementTree.Element
        The <text> XML element to parse.


    References
    ----------
    -  official matcher: https://github.com/spdx/spdx-license-matcher
    -  third-party matcher: https://github.com/MikeMoore63/spdx_matcher
    - Matching Guidelines: https://spdx.github.io/spdx-spec/v3.0.1/annexes/license-matching-guidelines-and-templates/
    """

    def __init__(self, text: ET.Element):
        self._text = text
        self._ns_uri = 'http://www.spdx.org/license'
        self._ns = {'': self._ns_uri}
        self._element_processor = {
            "titleText": self.title_text,
            "copyrightText": self.copyright_text,
            "standardLicenseHeader": self.standard_license_header,
            "list": self.list,
            "p": self.p,
            "br": self.br,
            "item": self.item,
            "bullet": self.bullet,
            "optional": self.optional,
            "alt": self.alt,
        }
        self._alt: dict = {}
        self._optionals: bool | list[bool] = True
        return

    def generate(
        self,
        alts: dict[str, str] | None = None,
        optionals: bool | list[bool] = True,
    ) -> Any:
        """Generate license full text and header.

        Parameters
        ----------
        alts : dict[str, int] | None, optional
            A dictionary specifying choices for <alt> elements. Keys are 'name' attributes,
            and values are the value to use.

        Returns
        -------
        The full text of the license, and the license header text, if present.
        """
        self._alt = alts or {}
        self._optionals = optionals
        return self.generate_full(self._text)

    def process(self, element: ET.Element) -> str:
        tag = self.clean_tag(element.tag)
        if tag not in self._element_processor:
            raise ValueError(f"Unsupported element: {tag}")
        processor = self._element_processor[tag]
        return processor(element)

    def get_alt(self, element: ET.Element) -> str:
        """Process an <alt> element by selecting the appropriate alternative based on `self._alt`.

        Parameters
        ----------
        element : xml.etree.ElementTree.Element
            The <alt> element.
        """
        name = element.get('name')
        match = element.get('match')
        if not name:
            raise ValueError("Alt element must have a 'name' attribute")
        if not match:
            raise ValueError("Alt element must have a 'match' attribute")
        text = self._alt.get(name)
        if not text:
            return element.text or ""
        if not _re.match(match, text):
            raise ValueError(f"Alt element '{name}' does not match '{match}'")
        return text

    def clean_tag(self, tag: str) -> str:
        """Strip the namespace URI from XML tag.

        Parameters
        ----------
        tag
            The XML tag with possible namespace.

        Returns
        -------
        The tag without namespace.
        """
        return tag.removeprefix(f'{{{self._ns_uri}}}')

    @staticmethod
    def clean_text(text: str) -> str:
        text_norm = _re.sub(r'\s+', ' ', text)
        if text_norm == " ":
            return ""
        return text_norm

    @staticmethod
    def element_has_text(element: ET.Element) -> bool:
        return bool(element.text and element.text.strip())

    @staticmethod
    def element_has_tail(element: ET.Element) -> bool:
        return bool(element.tail and element.tail.strip())

    def generate_full(self, text: ET.Element):
        raise NotImplementedError

    def generate_notice(self, sandard_license_header: ET.Element):
        raise NotImplementedError

    def title_text(self, element: ET.Element):
        raise NotImplementedError

    def copyright_text(self, element: ET.Element):
        raise NotImplementedError

    def standard_license_header(self, element: ET.Element):
        raise NotImplementedError

    def list(self, element: ET.Element):
        raise NotImplementedError

    def p(self, element: ET.Element):
        raise NotImplementedError

    def br(self, element: ET.Element):
        raise NotImplementedError

    def item(self, element: ET.Element):
        raise NotImplementedError

    def bullet(self, element: ET.Element):
        raise NotImplementedError

    def optional(self, element: ET.Element):
        raise NotImplementedError

    def alt(self, element: ET.Element):
        raise NotImplementedError


class SPDXLicenseTextPlain(SPDXLicenseText):
    """Plain-text generator for SPDX licenses.

    Parameters
    ----------
    text : xml.etree.ElementTree.Element
        The <text> XML element to parse.
    """

    def __init__(self, text: ET.Element):
        super().__init__(text)
        self._title: str | bool = True
        self._copyright: str | bool = False
        self._optionals: bool | list[bool] = True
        self._line_len: int = 88
        self._item_indent: int = 1
        self._item_spacing: int = 1
        self._current_list_nesting: int = 0
        self._list_indent: int = 4
        self._bullet: bool = True
        self._text_wrapper: _TextWrapper | None = None
        self._curr_bullet_len: int = 0
        self._title_centered: bool = True
        self._title_underline_full: bool = True
        self._title_underline: str = "="
        self._subtitle_underline: str = "–"
        self._count_optional = 0
        self._line_breaks = 2
        return

    def generate(
        self,
        title: str | bool = True,
        copyright_notice: str | bool = False,
        optionals: bool | list[bool] = True,
        alts: dict[str, str] | None = None,
        line_length: int = 88,
        list_indent: int = 0,
        item_indent: int = 1,
        item_spacing: int = 1,
        bullet: str | int | None = 1,
        title_centered: bool = False,
        title_underline: Literal["-", "=", "_", "*"] = "=",
        title_underline_full: bool = False,
        subtitle_underline: Literal["-", "=", "_", "*"] = "-",
        line_breaks: int = 2,
    ) -> str:
        """Generate plain-text license.

        Parameters
        ----------
        title
            Determines how to treat the license title, if any.
            Since the title is [optional](https://spdx.github.io/spdx-spec/v3.0.1/annexes/license-matching-guidelines-and-templates/#license-name-or-title)
            and not used in matching, it can be omitted or replaced with a custom title.
            If True, the title is included as-is. If False, the title is omitted.
            If a string, the title is replaced with the custom string, if a title is present.
        copyright_notice
            Determines how to treat the copyright notice, if any.
            Since the copyright notice is [optional](https://spdx.github.io/spdx-spec/v3.0.1/annexes/license-matching-guidelines-and-templates/#copyright-notice)
            and not used in matching, it can be omitted or replaced with a custom notice.
            If True, the notice is included as-is. If False, the notice is omitted.
            If a string, the notice is replaced with the custom string, if a notice is present.
        optionals : bool, optional
            Whether to include <optional> elements in the output, by default True.
        alts : dict[str, int] | None, optional
            A dictionary specifying choices for <alt> elements. Keys are 'name' attributes,
            and values are the value to use.
        line_length
            The maximum line length for the plain-text output.
        list_indent
            The number of spaces separating list items from the left margin.
        item_indent
            The number of spaces separating list items from the bullet character.
        item_spacing
            The number of newlines separating list items.
        bullet
            If `None`, the license's default bullet characters are used for list items.
            If a string, the specified character is used.
            If an integer, items are numbered starting from the specified number.
        title_centered
            Whether to center the title text.
        title_underline
            The character to use for underlining the title.
            Set to `None` to disable underlining.
        title_underline_full
            Whether to extend the underline to the full line length.
        subtitle_underline
            The character to use for underlining subtitles.
        line_breaks
            Number of newlines to add for each <br> element.

        Returns
        -------
        The plain-text version of the license
        plus the license header text, if present.
        """
        self._title = title
        self._copyright = copyright_notice
        self._optionals = optionals
        self._line_len = line_length
        self._text_wrapper = _TextWrapper(
            width=line_length,
            replace_whitespace=True,
            drop_whitespace=True,
            break_long_words=False,
            break_on_hyphens=False,
        )
        self._current_list_nesting = 0
        self._curr_bullet_len = 0
        self._count_optional = 0
        self._list_indent = list_indent
        self._item_indent = item_indent
        self._item_spacing = item_spacing
        self._bullet = bullet
        self._title_centered = title_centered
        self._title_underline = title_underline
        self._title_underline_full = title_underline_full
        self._subtitle_underline = subtitle_underline
        self._line_breaks = line_breaks
        fulltext = super().generate(alts=alts, optionals=optionals)
        return self.finalize(fulltext)

    def finalize(self, text: str | None) -> str:
        to_wrap_section_indices = []
        cleaned_sections = [[]]
        section_breaks = [0]
        in_break = False
        found_line = False
        for line in text.lstrip("\n").rstrip().splitlines():
            if not line.strip():
                section_breaks[-1] += 1
                in_break = True
                found_line = False
                continue
            if in_break:
                cleaned_sections.append([])
                section_breaks.append(0)
                in_break = False
            line_stripped = line.rstrip()
            if len(line_stripped) <= self._line_len:
                if not found_line:
                    cleaned_sections[-1].append(line_stripped)
                else:
                    line_indent = len(line) - len(line.lstrip())
                    last_line = cleaned_sections[-1][-1]
                    last_line_indent = len(last_line) - len(last_line.lstrip())
                    if last_line_indent == line_indent:
                        cleaned_sections[-1].append(line_stripped)
                    else:
                        cleaned_sections.append([line_stripped])
                        section_breaks.append(0)
                continue
            found_line = True
            if not cleaned_sections[-1]:
                cleaned_sections[-1].append(line_stripped)
            else:
                last_line = cleaned_sections[-1][-1]
                last_line_indent = len(last_line) - len(last_line.lstrip())
                current_line_indent = len(line) - len(line.lstrip())
                if last_line_indent == current_line_indent:
                    cleaned_sections[-1].append(line_stripped)
                else:
                    cleaned_sections.append([line_stripped])
                    section_breaks.append(0)
            to_wrap_section_indices.append(len(cleaned_sections) - 1)
        wrapped_sections = []
        for idx, section in enumerate(cleaned_sections):
            if idx in to_wrap_section_indices:
                wrapped_sections.append(self.wrap_text("\n".join(section)))
            else:
                wrapped_sections.append("\n".join(section))
            wrapped_sections.append("\n" * (section_breaks[idx] + 1))
        return f"{"".join(wrapped_sections).rstrip()}\n".replace(
            " .", "."
        ).replace(
            " ,", ","
        ).replace(
            " :", ":"
        ).replace(
            " ;", ";"
        ).replace(
            " )", ")"
        ).replace(
            " ]", "]"
        ).replace(
            " }", "}"
        ).replace(
            " !", "!"
        ).replace(
            " ?", "?"
        ).replace(
            "( ", "("
        ).replace(
            "[ ", "["
        ).replace(
            "{ ", "{"
        )

    def generate_full(self, text: ET.Element):
        return self.generic(text)

    def generic(
        self,
        element: ET.Element,
        return_list: bool = False,
    ) -> str | list[str]:
        """Recursively processes an XML element and its children.

        Parameters
        ----------
        element : xml.etree.ElementTree.Element
            The XML element to process.
        """
        # If the content begins on the same line as the tag, and has a leading space
        # (i.e. when first line has content with leading space)
        # then one space is added to the stripped text to preserve the leading space.
        # Similarly, if the content ends on the same line as the tag, and has a trailing space,
        # then one space is added to the end of the stripped text.

        # text_lines = text.splitlines()
        # space_start = bool(text_lines[0].strip() and text_lines[0].startswith(" ")) * " "
        # space_end = bool(text_lines[-1].strip() and text_lines[-1].endswith(" ")) * " "
        # space_normalized_text = _re.sub(r'\s+', ' ', text.strip())
        # space_normalized_text = f"{space_start}{space_normalized_text}{space_end}"
        # wrapped_text = self.wrap_text(space_normalized_text)
        # if space_normalized_text.endswith(" "):
        #     wrapped_text += " "
        # return wrapped_text

        out = []
        children = list(element)
        bullet_content = None
        for child_idx, child in enumerate(children):
            tag_name = self.clean_tag(child.tag)
            if tag_name not in self._element_processor:
                raise ValueError(f"Unsupported element: {tag_name}")
            content = self._element_processor[tag_name](child)
            if content:
                if not bullet_content:
                    out.append(content)
                else:
                    bullet_len = len(bullet_content)
                    bullet_content = None
                    self._curr_bullet_len -= bullet_len
                    content_lines = content.strip("\n").splitlines()
                    out.append(f"{content_lines[0].strip()}\n")
                    out.extend([f"{bullet_len * " "}{line}\n" for line in content_lines[1:]])
                    if self._subtitle_underline and len(content_lines) == 1:
                        num_chars = len(content_lines[0].strip()) + bullet_len
                        out.append(f"{self._subtitle_underline * num_chars}\n\n")
                    else:
                        out.append("\n\n")
            if tag_name == "bullet":
                # There is a bullet element outside of a list item.
                if self.element_has_tail(child):
                    if self._subtitle_underline:
                        num_chars = len(content.strip())
                        leading_spaces = (len(content) - len(content.lstrip(' '))) * " "
                        out.append(f"{leading_spaces}{self._subtitle_underline * num_chars}\n\n")
                else:
                    # The bullet has no text after it (example: CPL-1.0);
                    # Add the next element as the list item text.
                    bullet_content = content
                    self._curr_bullet_len += len(content)
        if self.element_has_text(element):
            out.insert(0, self.process_text(element.text))
        if self.element_has_tail(element):
            out.append(self.process_text(element.tail))
        if return_list:
            return out
        # full_raw = "".join([line.rstrip(" ") if line.strip() else "\n" for elem in out for line in elem.splitlines()])
        # paragraphs = [paragraph.strip() for paragraph in _re.split(r'\n\s*\n+', full_raw)]
        # processed = [self.wrap_text(paragraph) for paragraph in paragraphs]
        # return "\n\n".join(processed)
        return _re.sub(r'\n\s*\n\s*\n+', "\n\n", "".join(out)) + " "

    def standard_license_header(self, element: ET.Element):
        return self.generic(element)

    def title_text(self, element: ET.Element) -> str:
        """Process a <titleText> element."""
        if self._title is False:
            return ""
        title = self.generic(element) if self._title is True else self._title
        title_lines = []
        for line in title.splitlines():
            line = line.strip()
            if not line:
                continue
            if self._title_underline and all(char in ("-", "=", "_", "*") for char in line):
                continue
            if self._title_centered:
                line = line.center(self._line_len)
            title_lines.append(line)
        if self._title_underline:
            if self._title_underline_full:
                title_lines.append(self._title_underline * self._line_len)
            else:
                separator_line = self._title_underline * max(len(line) for line in title_lines)
                if self._title_centered:
                    separator_line = separator_line.center(self._line_len)
                title_lines.append(separator_line)
        title_lines.append("\n")
        return "\n".join(title_lines)

    def copyright_text(self, element: ET.Element) -> str:
        """Process a <copyrightText> element."""
        if self._copyright is False:
            return ""
        copyright_text = self.generic(element) if self._copyright is True else self._copyright
        return f"\n\n{copyright_text.strip()}\n\n"

    def optional(self, element: ET.Element) -> str:
        """
        Processes an <optional> element based on the include_optional flag.

        Parameters
        ----------
        element : xml.etree.ElementTree.Element
            The <optional> element.
        """
        include = self._optionals if isinstance(self._optionals, bool) else self._optionals[self._count_optional]
        self._count_optional += 1
        if not include:
            return self.process_text(element.tail or "")
        return self.generic(element)

    def list(self, elem: ET.Element) -> str:
        """
        Processes a <list> element containing <item> elements.

        Parameters
        ----------
        elem : xml.etree.ElementTree.Element
            The <list> element.
        """
        self._current_list_nesting += 1
        if elem.text and elem.text.strip():
            raise ValueError("List element should not have text content")
        if self._bullet:
            bullet_elems = elem.findall("./item/bullet", self._ns) + elem.findall("./item/p/bullet", self._ns)
            max_bullet_width = max([len(bullet.text.strip()) for bullet in bullet_elems], default=0)
        else:
            max_bullet_width = len(str(len(elem)))
        items = []
        for idx, child in enumerate(elem):
            tag = self.clean_tag(child.tag)
            if tag != 'item':
                raise ValueError(f"List element should only contain item elements, not {tag}")
            item_str = self.item(child, idx, max_bullet_width=max_bullet_width)
            item_str_indented = "\n".join(
                [f"{' ' * self._list_indent}{line}" for line in item_str.splitlines()])
            items.append(item_str_indented)
        self._current_list_nesting -= 1
        newlines = max(1, self._item_spacing + 1) * "\n"
        list_str = newlines.join(items)
        return f"{newlines}{list_str}{newlines}"

    def item(self, elem: ET.Element, idx: int, max_bullet_width: int) -> str:
        bullet_elems = elem.findall("./bullet", self._ns) + elem.findall("./p/bullet", self._ns)
        if len(bullet_elems) > 1:
            raise ValueError("Item element should contain at most one bullet element")
        if len(bullet_elems) == 1:
            bullet = bullet_elems[0].text.strip() if not self._bullet else (
                f"{idx + self._bullet}." if isinstance(self._bullet, int) else self._bullet
            )
            bullet_post_space = max_bullet_width + self._item_indent - len(bullet)
            bullet += bullet_post_space * " "
            subsequent_indent = len(bullet) * " "
        else:
            bullet = ""
            subsequent_indent = ""
        self._curr_bullet_len += len(bullet)
        content = []
        if elem.text:
            text = self.process_text(elem.text).lstrip()
            if text:
                content.append(text)
        for child in elem:
            tag = self.clean_tag(child.tag)
            if tag == 'bullet':
                if self.element_has_tail(child):
                    tail = self.process_text(child.tail)
                    content.append(tail)
            else:
                child_str = self.process(child)
                if child_str:
                    content.append(child_str.lstrip(" "))
            # if child.tail:
            #     tail = self.process_text(child.tail)
            #     if tail:
            #         needs_dedent = not content or content[-1].endswith("\n")
            #         content.append(tail.lstrip() if needs_dedent else tail)
        content_raw = "".join(content).strip()

        lines = content_raw.splitlines()
        wrapped = "\n".join(
            [f"{bullet}{lines[0] if lines else ""}"] + [f"{subsequent_indent}{line}" for line in lines[1:]]
        )
        self._curr_bullet_len -= len(bullet)
        return wrapped

    def p(self, element: ET.Element) -> str:
        """
        Processes a <p> element and appends its text to the output.

        Parameters
        ----------
        element : xml.etree.ElementTree.Element
            The <p> element.
        """
        out = [[]]
        if element.text:
            out[-1].append(element.text)
        for child in element:
            tag_name = self.clean_tag(child.tag)
            if tag_name not in self._element_processor:
                raise ValueError(f"Unsupported element: {tag_name}")
            if tag_name == "br":
                out.append([])
                if child.tail:
                    out[-1].append(child.tail)
            elif tag_name != "bullet":
                # Sometimes the <bullet> for <item> is placed inside a <p> element of that item.
                # Here we ignore the <bullet> element since `item()` will handle it.
                content = self._element_processor[tag_name](child)
                if content:
                    out[-1].append(content)
        if element.tail:
            out[-1].append(element.tail)

        paragraphs = []
        for paragraph_components in out:
            paragraph_raw = " ".join(paragraph_components)
            paragraph_normalized = _re.sub(r'\s+', ' ', paragraph_raw).strip()
            paragraphs.append(self.wrap_text(paragraph_normalized))
        return f"\n\n{("\n" * self._line_breaks).join(paragraphs)}\n\n"

    def alt(self, element: ET.Element) -> str:
        """Process an <alt> element by selecting the appropriate alternative based on `self._alt`.

        Parameters
        ----------
        element : xml.etree.ElementTree.Element
            The <alt> element.
        """
        element.text = super().get_alt(element)
        return self.p(element)

    def bullet(self, element: ET.Element) -> str:
        # This is only called when the bullet is outside of a list item.
        if list(element):
            raise ValueError("Bullet element should not have children")
        if not self.element_has_text(element):
            raise ValueError("Bullet element should have text content")
        bullet = f"{element.text.strip()}{" " * self._item_indent}"
        item = f"{bullet}{element.tail.strip()}" if self.element_has_tail(element) else bullet
        return f"\n{self.process_text(item)}\n"

    def br(self, element: ET.Element) -> str:
        if self.element_has_text(element):
            raise ValueError("BR element should not have text content")
        if list(element):
            raise ValueError("BR element should not have children")
        if self.element_has_tail(element):
            tail = self.process_text(element.tail)
        else:
            tail = ""
        return f"{"\n" * self._line_breaks}{tail} "

    def process_text(self, text: str) -> str:
        space_normalized_text = _re.sub(r'\s+', ' ', text.strip())
        wrapped_text = self.wrap_text(space_normalized_text)
        return f"{wrapped_text} "

    def wrap_text(self, text: str) -> str:
        """Wrap text to the specified line length, preserving indentation.

        Parameters
        ----------
        text : str
            The text to wrap.
        """
        if self._current_list_nesting:
            extra_width = (self._current_list_nesting * self._list_indent) + self._curr_bullet_len
        else:
            extra_width = 0
        self._text_wrapper.width = self._line_len - extra_width
        wrapped = self._text_wrapper.fill(text)
        return wrapped


class SPDXLicenseTextMD(SPDXLicenseText):
    """Parses the <text> element from an SPDX license XML and generates a Markdown version of the license.

    Parameters
    ----------
    text : xml.etree.ElementTree.Element
        The <text> XML element to parse.
    """

    def __init__(self, text: ET.Element):
        super().__init__(text)
        return

    def generate(
        self,
        alts: dict[str, str] | None = None,
    ) -> tuple[_mdit.Document, _mdit.Document | None]:
        """Generate Markdown license.

        Returns
        -------
        The Markdown version of the license.
        """
        return super().generate(alts)



from __future__ import annotations as annotations

from typing import TYPE_CHECKING as TYPE_CHECKING
import json as _json

from licenseman.spdx.exception import SPDXLicenseException as _SPDXLicenseException

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Sequence, Generator
    from licenseman.spdx.exception_list import SPDXExceptionList


class SPDXLicenseExceptionDB:

    def __init__(
        self,
        exception_list: SPDXExceptionList,
        db_path: Path,
        in_memory: bool = False,
        verify: bool = True,
        exceptions: dict[str, _SPDXLicenseException] | None = None,
    ):
        self._exception_list = exception_list
        self._db_path = db_path
        self._in_memory = in_memory
        self._verify = verify
        self._exceptions: dict[str, _SPDXLicenseException] = exceptions or {}
        if in_memory:
            self.load(verify=verify)
        elif verify:
            for _ in self.get(verify=True):
                pass
        return

    def load(self, exception_ids: Sequence[str] | None = None, verify: bool | None = None):
        exception_ids = exception_ids or self._exception_list.ids
        for exception_id, exception_data in zip(exception_ids, self.get(exception_ids, verify=verify)):
            self._exceptions[exception_id] = exception_data
        return

    def get(self, exception_ids: Sequence[str] | None = None, verify: bool | None = None) -> Generator[_SPDXLicenseException]:
        for exception_id in exception_ids or self._exception_list.ids:
            if exception_id in self._exceptions:
                yield self._exceptions[exception_id]
            else:
                with open(self._db_path / f"{exception_id}.json") as f:
                    data = _json.load(f)
                yield _SPDXLicenseException(
                    data,
                    verify=verify if verify is not None else self._verify,
                )

    def alts(self, exception_ids: Sequence[str] | None = None):
        exception_ids = exception_ids or self._exception_list.ids
        alts: dict[str, list[dict[str, str]]] = {}
        for exception in self.get(exception_ids):
            for alt_name, alt_data in exception.alts.items():
                alts.setdefault(alt_name, []).append({"id": exception.id, **alt_data})
        return alts

    def __getitem__(self, exception_id: str) -> _SPDXLicenseException:
        return self.get([exception_id]).__next__()

    def __contains__(self, exception_id: str) -> bool:
        return exception_id in self._exception_list

